// **************************************************************************
// test_ipclayer.cc,v 1.6 2007/01/30 18:53:31 cnepveu Exp
//
// Copyright (c) 2007 Hexago Inc. All rights reserved.
// 
//   LICENSE NOTICE: You may use and modify this source code only if you
//   have executed a valid license agreement with Hexago Inc. granting
//   you the right to do so, the said license agreement governing such
//   use and modifications.   Copyright or other intellectual property
//   notices are not to be removed from the source code.
//
// Description:
//   IPC Layer tester.
//
// Author: Charles Nepveu
//
// Creation Date: November 2006
// __________________________________________________________________________
// **************************************************************************
#include <gw6cmessaging/servent.h>
#include <gw6cmessaging/pipeclient.h>
#include <gw6cmessaging/pipeserver.h>
#include <gw6cmessaging/threadwrapper.h>
#include <iostream>
#include <iomanip>
using namespace std;
using namespace gw6cmessaging;
#include <stdlib.h>
#include <time.h>


#define CMDLINEOPT_STR_SERVER       "--server"
#define CMDLINEOPT_STR_CLIENT       "--client"

#define PIPE_NAME                   "\\\\.\\pipe\\gw6c-test"

#ifdef WIN32
#include <windows.h>
#define SLEEP(X)  Sleep(X)
#else
#include <unistd.h>
#define SLEEP(X)  usleep((X) * 1000)
#endif


// --------------------------------------------------------------------------
class SenderThread: public ThreadWrapper
{
private:
  IPCServent*   m_pIPCServent;
  Servent*      m_pServent;

public:
  SenderThread(IPCServent* pIPCServent, Servent* pServent) : 
      ThreadWrapper(), m_pIPCServent(pIPCServent), m_pServent(pServent) {};
  virtual ~SenderThread() {};
protected:
  void Work( void );
};


// --------------------------------------------------------------------------
class ReceiverThread: public ThreadWrapper
{
private:
  IPCServent*   m_pIPCServent;
  Servent*      m_pServent;

public:
  ReceiverThread(IPCServent* pIPCServent, Servent* pServent) : 
     ThreadWrapper(), m_pIPCServent(pIPCServent), m_pServent(pServent)  {};
  virtual ~ReceiverThread() {};
protected:
  void Work( void );
};


// --------------------------------------------------------------------------
// Global boolean.
bool gIsServer;


// --------------------------------------------------------------------------
// Prints the program usage
void PrintUsage( const char* szAppName )
{
  cout << "GW6C Messaging IPC layer tester." << endl;
  cout << "Start 2 instances of this program (one server and one client)." << endl;
  cout << endl << "Usage: " << szAppName << " <--server|--client>" << endl;
  cout << "  --server:   Starts a SERVER instance of the IPC layer." << endl;
  cout << "  --client:   Starts a CLIENT instance of the IPC layer." << endl << endl;
}


// --------------------------------------------------------------------------
// Parse the command-line option(s).
void ParseOptions( int argc, char* argv[] )
{
  // Check number of arguments passed.
  if( argc != 2 )
  {
    PrintUsage( argv[0] );
    exit(1);
  }

  // Check if this instance is server.
  if( strcmp( argv[1], CMDLINEOPT_STR_SERVER ) == 0 )
  {
    gIsServer = true;
    return;
  }

  // Check if this instance is client.
  if( strcmp( argv[1], CMDLINEOPT_STR_CLIENT ) == 0 )
  {
    gIsServer = false;
    return;
  }

  // Bad option.
  PrintUsage( argv[0] );
  exit(1);
}

// --------------------------------------------------------------------------
void SenderThread::Work( void )
{
  unsigned char data[] = "";
  bool bCanRead;
  size_t nSent;
  error_t retCode;
  unsigned int i = 0;


  char c = (gIsServer) ? 'S' : 'C';
  memset( data, c, sizeof(data) );

  while( !ShouldStop() )
  {
    // Verify if we can read data
    if( (retCode=m_pServent->CanRead( bCanRead )) != GW6CM_UIS__NOERROR )
    {
      // Error peeking on the pipe.
      cout << "Error peeking. Error code: 0x" << setbase(16) << retCode << setbase(10) << endl;
#ifdef WIN32
      cout << "GetLastError() returned: " << GetLastError() << endl;
#endif
      exit( 1 );
    }
    if( bCanRead )
    {
      SLEEP( 25 );     // Wait for data to be read.
      continue;
    }

    cout << "Sending message #" << i << " of " << sizeof(data) << " bytes...";

    retCode = m_pServent->WriteData( (void*)data, sizeof(data), nSent );
    if( retCode != GW6CM_UIS__NOERROR )
    {
      // An error occured.
      cout << "Error sending data. Error code: 0x" << setbase(16) << retCode << endl;
#ifdef WIN32
      cout << "GetLastError() returned: " << setbase(10) << GetLastError() << endl;
#endif
      exit( 1 );
    }

    if( nSent != sizeof(data) )
      cout << "Failed to send all data. Sent " << nSent << " bytes of data!" << endl;
    else
      cout << "SUCCESS!" << endl << "-> Sent all " << nSent << " bytes of data!" << endl;

    SLEEP( rand() % 2000 );
    i++;
  }
}


// --------------------------------------------------------------------------
void ReceiverThread::Work( void )
{
  unsigned char recvBuffer[64260];
  size_t nRecv;
  error_t retCode;
  bool bCanRead;


  while( !ShouldStop() )
  {
    // Verify if we can read data
    if( (retCode=m_pServent->CanRead( bCanRead )) != GW6CM_UIS__NOERROR )
    {
      // Error peeking on the pipe.
      cout << "Error peeking. Error code: 0x" << setbase(16) << retCode << setbase(10) << endl;
#ifdef WIN32
      cout << "GetLastError() returned: " << GetLastError() << endl;
#endif
      exit( 1 );
    }
    if( !bCanRead )
    {
      SLEEP( 25 );     // Wait for data to be available.
      continue;
    }
    cout << "Receiving message of " << nRecv << " raw bytes...";

    // Read the data.
    retCode = m_pServent->ReadData( (void*)recvBuffer, sizeof(recvBuffer), nRecv );
    if( retCode != GW6CM_UIS__NOERROR )
    {
      // An error occured.
      cout << "Error sending data. Error code: 0x" << setbase(16) << retCode << endl;
#ifdef WIN32
      cout << "GetLastError() returned: " << setbase(10) << GetLastError() << endl;
#endif
      exit( 1 );
    }

    cout << "SUCCESS!" << endl << "-> Received " << nRecv << " bytes of user data!" << endl;
    cout << recvBuffer << endl << endl;
  }

}


// --------------------------------------------------------------------------
void StartServer( void )
{
  PipeServer* pServer = new PipeServer( PIPE_NAME );
  Servent*    pServent = new Servent();
  SenderThread    st(pServer, pServent);
  ReceiverThread  rt(pServer, pServent);
  int nRetCode;


  // Initialize Servent oject.
  if( (nRetCode = pServent->Initialize( pServer )) != 0 )
  {
    cout << "Failed to Initialize server Servent object. Error code: " << nRetCode << endl;
    delete pServent;
    exit(1);
  }
  cout << "------------------------------------" << endl;
  cout << "   Server -> Client has connected." << endl;
  cout << "------------------------------------" << endl;


  // Send & receive Data until end of time.
  st.Run();
  rt.Run();

  SERVENT_INFO info;
  for(;;)
  {
    SLEEP(10000);
    pServent->GetServentInfo( &info );
    cout << "Statistics: Server sent " << info. nTtlBytesWritten << 
         " bytes and received " << info. nTtlBytesRead << " bytes." << endl;
  }
}


// --------------------------------------------------------------------------
void StartClient( void )
{
  PipeClient* pClient = new PipeClient( PIPE_NAME );
  Servent*    pServent = new Servent();
  SenderThread    st(pClient, pServent);
  ReceiverThread  rt(pClient, pServent);
  int nRetCode;


  // Initialize Servent oject.
  if( (nRetCode = pServent->Initialize( pClient )) != 0 )
  {
    cout << "Failed to Initialize server Servent object. Error code: " << nRetCode << endl;
    delete pServent;
    exit(1);
  }
  cout << "------------------------------------" << endl;
  cout << "   Client -> Connected to server." << endl;
  cout << "------------------------------------" << endl;


  // Send & receive Data until end of time.
  st.Run();
  rt.Run();

  SERVENT_INFO info;
  for(;;)
  {
    SLEEP(10000);
    pServent->GetServentInfo( &info );
    cout << "Statistics: Client sent " << info. nTtlBytesWritten << 
         " bytes and received " << info. nTtlBytesRead << " bytes." << endl;
  }
}


// --------------------------------------------------------------------------
// MAIN PROGRAM ENTRY
// --------------------------------------------------------------------------
int main( int argc, char* argv[] )
{
  ParseOptions( argc, argv );

  srand( time( NULL ) );

  if( gIsServer )
    StartServer();
  else
    StartClient();


  return 0;
}
