// **************************************************************************
// test_messaginglayer.cc,v 1.10 2007/01/30 18:53:31 cnepveu Exp
//
// Copyright (c) 2007 Hexago Inc. All rights reserved.
// 
//   LICENSE NOTICE: You may use and modify this source code only if you
//   have executed a valid license agreement with Hexago Inc. granting
//   you the right to do so, the said license agreement governing such
//   use and modifications.   Copyright or other intellectual property
//   notices are not to be removed from the source code.
//
// Description:
//   IPC Layer tester.
//
// Author: Charles Nepveu
//
// Creation Date: November 2006
// __________________________________________________________________________
// **************************************************************************
#include <gw6cmessaging/communicationsmgr.h>
#include <gw6cmessaging/clientmsgsender.h>
#include <gw6cmessaging/clientmsgtranslator.h>
#include <gw6cmessaging/servermsgsender.h>
#include <gw6cmessaging/servermsgtranslator.h>
#include <stdlib.h>
#include <time.h>
#include <iostream>
#include <iomanip>
using namespace std;
using namespace gw6cmessaging;


#ifdef WIN32
#include <windows.h>
#define SLEEP(X)  Sleep(X)
#undef PostMessage
#else
#include <unistd.h>
#define SLEEP(X)  usleep((X) * 1000)
#endif

#define CMDLINEOPT_STR_SERVER       "--server"
#define CMDLINEOPT_STR_CLIENT       "--client"


// --------------------------------------------------------------------------
class ClientImplTest: public ClientMsgSender, public ClientMsgTranslator
{
public:
  CommunicationsManager* pManager;


  ClientImplTest( void ) : ClientMsgSender(), ClientMsgTranslator() 
  {
    pManager = new CommunicationsManager( CLIENT_MANAGER, this );
    pManager->Run();
  }
  virtual           ~ClientImplTest       ( void ) {}

  // Virtual overrides from ClientMsgTranslator
  error_t Recv_StatusInfo( const Gw6cStatusInfo* aStatusInfo )
  {
    cout << "CLIENT: Received status info from server!!" << endl;
    return GW6CM_UIS__NOERROR;
  }

  error_t Recv_TunnelInfo( const Gw6cTunnelInfo* aTunnelInfo )
  {
    cout << "CLIENT: Received tunnel info from server!!" << endl;
    return GW6CM_UIS__NOERROR;
  }

  error_t Recv_BrokerList( const Gw6cBrokerList* aBrokerList )
  {
    cout << "CLIENT: Received broker list from server!!" << endl;
    return GW6CM_UIS__NOERROR;
  }

  // Virtual overrides from ClientMsgSender
  void PostMessage( Message* pMsg )
  {
    MessageSender* pSender = (MessageSender*)pManager;
    pSender->PostMessage( pMsg );
  }

};


// --------------------------------------------------------------------------
class ServerImplTest: public ServerMsgSender, public ServerMsgTranslator
{
public:
  CommunicationsManager* pManager;


  ServerImplTest( void ) : ServerMsgSender(), ServerMsgTranslator()
  {
    pManager = new CommunicationsManager( SERVER_MANAGER, this );
    pManager->Run();
  }

  virtual ~ServerImplTest( void ) 
  {}

  // Virtual overrides from ServerMsgTranslator
  error_t Recv_StatusInfoRequest( /*custom params TBD*/ )
  {
    Gw6cStatusInfo statusInfo;
    statusInfo.eStatus = (Gw6cCliStatus) (rand() % 4);
    statusInfo.nStatus = GW6CM_UIS__NOERROR;
    cout << "SERVER: Received status info request from client. Sending requested info..." << endl;
    Send_StatusInfo(&statusInfo);
    return GW6CM_UIS__NOERROR;
  }

  error_t Recv_TunnelInfoRequest( /*custom params TBD*/ )
  {
    Gw6cTunnelInfo tunnelInfo;
    memset(&tunnelInfo, 0x00, sizeof(Gw6cTunnelInfo));
    cout << "SERVER: Received tunnel info request from client. Sending requested info..." << endl;
    Send_TunnelInfo(&tunnelInfo);
    return GW6CM_UIS__NOERROR;
  }

  error_t Recv_BrokerListRequest( /*custom params TBD*/ )
  {
    Gw6cBrokerList brokerList;
    memset(&brokerList, 0x00, sizeof(Gw6cBrokerList));
    cout << "SERVER: Received broker list request from client. Sending requested info..." << endl;
    Send_BrokerList(&brokerList);
    return GW6CM_UIS__NOERROR;
  }

  // Virtual overrides from ClientMsgSender
  void PostMessage( Message* pMsg )
  {
    MessageSender* pSender = (MessageSender*)pManager;
    pSender->PostMessage( pMsg );
  }
};


// --------------------------------------------------------------------------
// Global boolean.
bool gIsServer;


// --------------------------------------------------------------------------
// Prints the program usage
void PrintUsage( const char* szAppName )
{
  cout << "GW6C Messaging layer tester." << endl;
  cout << "Start 2 instances of this program (one server and one client)." << endl;
  cout << endl << "Usage: " << szAppName << " <--server|--client>" << endl;
  cout << "  --server:   Starts a server instance of the messaging layer." << endl;
  cout << "  --client:   Starts a client instance of the messaging layer." << endl << endl;
}


// --------------------------------------------------------------------------
// Parse the command-line option(s).
void ParseOptions( int argc, char* argv[] )
{
  // Check number of arguments passed.
  if( argc != 2 )
  {
    PrintUsage( argv[0] );
    exit(1);
  }

  // Check if this instance is server.
  if( strcmp( argv[1], CMDLINEOPT_STR_SERVER ) == 0 )
  {
    gIsServer = true;
    return;
  }

  // Check if this instance is client.
  if( strcmp( argv[1], CMDLINEOPT_STR_CLIENT ) == 0 )
  {
    gIsServer = false;
    return;
  }

  // Bad option.
  PrintUsage( argv[0] );
  exit(1);
}


// --------------------------------------------------------------------------
void StartServer( void )
{
  ServerImplTest Messenger;
  CommunicationsManager::MANAGER_STATISTICS stats;

  // Sleep infinately. Only serve requests.
  for(;;)
  {
    SLEEP(10000);

    // Display communications manager current state:
    cout << "**  Communications manager state: ";
    switch( Messenger.pManager->GetState() )
    {
    case STATE_DISCONNECTED:      cout << "Disconnected." << endl; break;
    case STATE_PENDINGCONNECTION: cout << "Pending connection." << endl; break;
    case STATE_CONNECTED:         cout << "Connected." << endl; break;
    case STATE_FATALERROR:        cout << "FATAL ERROR." << endl; break;
    }

    // Display communications manager statistics:
    Messenger.pManager->GetStatistics( &stats );
    cout << "**  Messages [processed, sent, queued]: [" << stats.nMsgProcessed 
         << ", " << stats.nMsgSent << ", " << stats.nMsgQueued << "]" << endl;
    cout << "**  Bytes [sent, received]: [" << stats.ServentInfo.nTtlBytesWritten 
         << ", " << stats.ServentInfo.nTtlBytesRead << "]" << endl;
  }
}


// --------------------------------------------------------------------------
void StartClient( void )
{
  ClientImplTest Messenger;

  for(;;)
  {
    SLEEP((rand() % 20) + 10);    // Sleep 1 to 3 seconds.

    // Issue a random request.
    cout << "Issuing a random request: ";
    switch( rand() % 3 )
    {
    case 0:   
      cout << "STATUS INFO" << endl;
      Messenger.Send_StatusInfoRequest();
      break;
    case 1:   
      cout << "TUNNEL INFO" << endl;
      Messenger.Send_TunnelInfoRequest();
      break;
    case 2:   
      cout << "BROKER LIST" << endl;
      Messenger.Send_BrokerListRequest();
      break;
    }
  }
}


// --------------------------------------------------------------------------
// MAIN PROGRAM ENTRY
// --------------------------------------------------------------------------
int main( int argc, char* argv[] )
{
  ParseOptions( argc, argv );

  if( gIsServer )
    StartServer();
  else
    StartClient();


  return 0;
}
